#!/usr/bin/python
#-------------------------------python------------------------fract.py--#
#                                                                       #
#                             Fractions                                 #
#                                                                       #
# Copyright (c) 2013 Daniele Zambelli
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#--Daniele Zambelli--------------GPL-----------------------------2013---#

"""
Frazioni
"""

from __future__ import print_function, division
from symbbase import MultiOutputBase

VERSION = '0.8'

INT_TYPE = type(0)

####
# Tools
###


def mcd(num0, num1):
    """Massimo Comune Divisore tra num0 e num1."""
    try:
        if num0 & -1 and num1 & -1:
            num0 = abs(num0)
            num1 = abs(num1)
            while num1:
                num0, num1 = num1, num0 % num1
        else:
            num0 = 0
    except TypeError:   # num0 or num1 is not int
        num0 = 0
    return num0


def segno_no_piu(num):
    """Return '-' if < 0 or '0' if >= 0."""
    if num < 0:
        return "-"
    else:
        return ""


def sqrt(fra):
    """Restituisce la radice quadrata di una frazione.
>>> print(sqrt(Fraction(9, 4)))
\\frac{3}{2}
>>> print(sqrt(Fraction(16, 3)))
\\frac{4}{1.73205080757}
        """
    if fra.num:
        num = (fra.num) ** .5
        den = (fra.den) ** .5
        inum = int(round(num))
        iden = int(round(den))
        if abs(fra.num - inum) < 1e-8:
            fra.num = inum
        if abs(fra.den - iden) < 1e-8:
            fra.den = iden
    else:
        fra.num = 0
        fra.den = 1
    return Fraction(num, den)

####
# Classes
###


class FractionError(Exception):
    """Fraction error."""


class Fraction(MultiOutputBase):
    """La classe Fraction implementa il confronto,
    la riduzione ai minimi termini
    e la visualizzazione delle frazioni nei formati:
    - Open document Math
    - LaTeX
    - text."""

    def __init__(self, num=0, den=1):
        """
>>> a = Fraction(0)
>>> b = Fraction(15, 20)
>>> c = Fraction(1, 2)
>>> d = Fraction(-3, 2)
>>> e = Fraction(1, -3)
>>> f = Fraction(6, 2)
>>> g = Fraction(121, 225)
>>> Fraction(4, 0)
Traceback (most recent call last):
    ...
ZeroDivisionError: Fraction con denominatore uguale a zero
>>> print(a)
0
>>> print(b)
\\frac{3}{4}
>>> print(c)
\\frac{1}{2}
>>> print(d)
\\frac{-3}{2}
>>> print(e)
\\frac{-1}{3}
>>> print(f)
3
>>> print(g)
\\frac{121}{225}
>>> a.out_format = ODT
>>> print(a)
0
>>> print(b)
{3 over 4}
>>> print(c)
{1 over 2}
>>> print(d)
-{3 over 2}
>>> print(e)
-{1 over 3}
>>> print(f)
3
>>> print(g)
{121 over 225}
>>> a.out_format = PY
>>> print(a)
0
>>> print(b)
3/4
>>> print(c)
1/2
>>> print(d)
-3/2
>>> print(e)
-1/3
>>> print(f)
3
>>> print(g)
121/225
>>> a.out_format = TEX
        """
        if den == 0:
            raise ZeroDivisionError('Fraction con denominatore uguale a zero')
        elif den < 0:
            den = -den
            num = -num
#        if type(num) == None:
#            self.num = num.num
#            self.den = num.den*den
#        else:
#            self.num = num
#            self.den = den
        self.num = num
        self.den = den
        self.riduce()

    def __str__(self):
        """str."""
        return [self._str_odt,
                self._str_tex,
                self._str_py][self.out_format]()

    def _str_odt(self):
        """.odt string."""
        if self.den == 1:
            return str(self.num)
        if self.num == 0:
            return '0'
        return r'{0}{{{1} over {2}}}'.format(segno_no_piu(self.num),
                                             abs(self.num), self.den)

    def _str_tex(self):
        """.tex string."""
        if self.den == 1:
            return str(self.num)
        if self.num == 0:
            return '0'
        return r'\frac{{{0}}}{{{1}}}'.format(self.num, self.den)

    def _str_py(self):
        """.txt string."""
        if self.den == 1:
            return str(self.num)
        if self.num == 0:
            return '0'
        return r'{0}/{1}'.format(self.num, self.den)

    def riduce(self):
        """Riduce ai minimi termini la frazione e
        la restituisce come risultato."""
        if self.num:
            inum = int(round(self.num))
            iden = int(round(self.den))
            if abs(self.num - inum) < 1e-8:
                self.num = inum
            if abs(self.den - iden) < 1e-8:
                self.den = iden
            div = mcd(self.num, self.den)
            if div > 1:
                self.num //= div
                self.den //= div
        else:
            self.num = 0
            self.den = 1

    def __eq__(self, other):
        """Confronta due frazioni o una frazione e un intero: ==.
>>> a = Fraction(15, 20)
>>> b = Fraction(3, 2)
>>> c = Fraction(9, 12)
>>> print(a, "==", b, a == b)
\\frac{3}{4} == \\frac{3}{2} False
>>> print(a, "==", a, a == a)
\\frac{3}{4} == \\frac{3}{4} True
>>> print(a, "==", c, a == c)
\\frac{3}{4} == \\frac{3}{4} True
>>> print(a, "==", 3, a == 3)
\\frac{3}{4} == 3 False
>>> print(3, "==", a, 3 == a)
3 == \\frac{3}{4} False
        """
        if other == None:
            return False
        try:
            return self.num * other.den == self.den * other.num
        except AttributeError:
            try:
                return  self.num == self.den * other
            except AttributeError:
                raise FractionError('{0} not comparable with Fraction'
                                    ''.format(other))

    def __ne__(self, other):
        """Confronta due frazioni o una frazione e un intero: !=.
>>> a = Fraction(15, 20)
>>> b = Fraction(3, 2)
>>> c = Fraction(9, 12)
>>> print(a, "!=", b, a != b)
\\frac{3}{4} != \\frac{3}{2} True
>>> print(a, "!=", a, a != a)
\\frac{3}{4} != \\frac{3}{4} False
>>> print(a, "!=", c, a != c)
\\frac{3}{4} != \\frac{3}{4} False
>>> print(a, "!=", 3, a != 3)
\\frac{3}{4} != 3 True
>>> print(3, "!=", a, 3 != a)
3 != \\frac{3}{4} True
        """
        return not self.__eq__(other)

    def __gt__(self, other):
        """Confronta due frazioni o una frazione e un intero: >.
>>> a = Fraction(15, 20)
>>> b = Fraction(3, 2)
>>> c = Fraction(9, 12)
>>> print(a, ">", a, a > a)
\\frac{3}{4} > \\frac{3}{4} False
>>> print(a, ">", b, a > b)
\\frac{3}{4} > \\frac{3}{2} False
>>> print(a, ">", 2, a > 2)
\\frac{3}{4} > 2 False
>>> print(2, ">", b, 2 > b)
2 > \\frac{3}{2} True
        """
        try:
            return self.num * other.den > self.den * other.num
        except AttributeError:
            try:
                return  self.num > self.den * other
            except AttributeError:
                raise FractionError('{0} not comparable with Fraction'
                                    ''.format(other))

    def __lt__(self, other):
        """Confronta due frazioni o una frazione e un intero: <.
>>> a = Fraction(15, 20)
>>> b = Fraction(3, 2)
>>> c = Fraction(9, 12)
>>> print(a, "<", a, a < a)
\\frac{3}{4} < \\frac{3}{4} False
>>> print(a, "<", b, a < b)
\\frac{3}{4} < \\frac{3}{2} True
>>> print(a, "<", 2, a < 2)
\\frac{3}{4} < 2 True
>>> print(2, "<", b, 2 < b)
2 < \\frac{3}{2} False
        """
        return not self.__ge__(other)

    def __ge__(self, other):
        """Confronta due frazioni o una frazione e un intero: >=.
>>> a = Fraction(15, 20)
>>> b = Fraction(3, 2)
>>> c = Fraction(9, 12)
>>> print(a, ">=", a, a >= a)
\\frac{3}{4} >= \\frac{3}{4} True
>>> print(a, ">=", b, a >= b)
\\frac{3}{4} >= \\frac{3}{2} False
>>> print(a, ">=", 2, a >= 2)
\\frac{3}{4} >= 2 False
>>> print(2, ">=", b, 2 >= b)
2 >= \\frac{3}{2} True
        """
        try:
            return self.num * other.den >= self.den * other.num
        except AttributeError:
            try:
                return  self.num >= self.den * other
            except AttributeError:
                raise FractionError('{0} not comparable with Fraction'
                                    ''.format(other))

    def __le__(self, other):
        """Confronta due frazioni o una frazione e un intero: <=.
>>> a = Fraction(15, 20)
>>> b = Fraction(3, 2)
>>> c = Fraction(9, 12)
>>> print(a, "<=", a, a <= a)
\\frac{3}{4} <= \\frac{3}{4} True
>>> print(a, "<=", b, a <= b)
\\frac{3}{4} <= \\frac{3}{2} True
>>> print(a, "<=", 2, a <= 2)
\\frac{3}{4} <= 2 True
>>> print(2, "<=", b, 2 <= b)
2 <= \\frac{3}{2} False
        """
        return not self.__gt__(other)

    def __neg__(self):
        """Restituisce l'opposto di una frazione.
>>> print(-Fraction(48, 28))
\\frac{-12}{7}
        """
        return Fraction(- self.num, self.den)

    def __add__(self, other):
        """Restituisce la somma di due frazioni o
        di una frazione e un intero.
>>> print(Fraction(49, 28) + Fraction(24, 10))
\\frac{83}{20}
>>> print(Fraction(48, 28) + 3)
\\frac{33}{7}
>>> print(3 + Fraction(48, 28))
\\frac{33}{7}
        """
        try:
            return Fraction(self.num * other.den + self.den * other.num,
                            self.den * other.den)
        except AttributeError:
            try:
                other = Fraction(other)
                return Fraction(self.num * other.den + self.den * other.num,
                                self.den * other.den)
            except AttributeError:
                raise FractionError("{0} can not be converted".format(other))

    def __sub__(self, other):
        """Restituisce la differenza di due frazioni o
        di una frazione e un intero.
>>> print(Fraction(2, 6) - Fraction(3, 5))
\\frac{-4}{15}
>>> print(Fraction(48, 28) - 2)
\\frac{-2}{7}
>>> print(2 - Fraction(48, 28))
\\frac{2}{7}
        """
        try:
            return Fraction(self.num * other.den - self.den * other.num,
                            self.den * other.den)
        except AttributeError:
            try:
                other = Fraction(other)
                return Fraction(self.num * other.den - self.den * other.num,
                                self.den * other.den)
            except AttributeError:
                raise FractionError("{0} can not be converted".format(other))

    def __mul__(self, other):
        """Restituisce il prodotto di due frazioni o
        di una frazione e un intero.
>>> print(Fraction(21, 12) * Fraction(12, 5))
\\frac{21}{5}
>>> print(Fraction(16, 96) * 2)
\\frac{1}{3}
>>> print(2 * Fraction(16, 96))
\\frac{1}{3}
        """
        try:
            return Fraction(self.num * other.num, self.den * other.den)
        except AttributeError:
            try:
                other = Fraction(other)
                return Fraction(self.num * other.num, self.den * other.den)
            except AttributeError:
                raise FractionError("{0} can not be converted".format(other))

    def reciproco(self):
        """Restituisce il reciproco della frazione o
        di una frazione e un intero"""
        return Fraction(self.den, self.num)

    def __truediv__(self, other):
        """Restituisce il quoziente di due frazioni.
>>> print(Fraction(21, 12) / Fraction(5, 12))
\\frac{21}{5}
>>> print(Fraction(25, 15) / 2)
\\frac{5}{6}
>>> print(2 / Fraction(25, 15))
\\frac{6}{5}
        """
        try:
            return Fraction(self.num * other.den, self.den * other.num)
        except AttributeError:
            try:
                other = Fraction(other)
                return Fraction(self.num * other.den, self.den * other.num)
            except AttributeError:
                raise FractionError("{0} can not be converted".format(other))

#    def __pow__(self, exp):
#        """Restituisce la potenza di una frazione"""
#        return Fraction(self.num**exp, self.den**exp)

    def __pow__(self, exp):
        """Restituisce la potenza di una frazione.
>>> print(Fraction(9, 4)**2)
\\frac{81}{16}
>>> print(Fraction(2, 3)**5)
\\frac{32}{243}
        """
        if self.num:
            num = (self.num) ** exp
            den = (self.den) ** exp
            inum = int(round(num))
            iden = int(round(den))
            if abs(self.num - inum) < 1e-8:
                self.num = inum
            if abs(self.den - iden) < 1e-8:
                self.den = iden
        else:
            self.num = 0
            self.den = 1
        return Fraction(num, den)

    def __radd__(self, other):
        """Restituisce la somma di due frazioni o
        di una frazione e un intero"""
        return self.__add__(other)

    def __rsub__(self, other):
        """Restituisce la somma di due frazioni o
        di una frazione e un intero"""
        return - self.__sub__(other)

    def __rmul__(self, other):
        """Restituisce la somma di due frazioni o
        di una frazione e un intero"""
        return self.__mul__(other)

    def __rtruediv__(self, other):
        """Restituisce la somma di due frazioni o
        di una frazione e un intero"""
        return self.__truediv__(other).reciproco()

    def __int__(self):
        """Restituisce il valore intero della frazione"""
        return int(self.num) / int(self.den)

    def __float__(self):
        """Restituisce il valore decimale della frazione"""
        return float(self.num) / self.den

    def __abs__(self):
        """Restituisce il valore assoluto della frazione"""
        return Fraction(abs(self.num), self.den)

#    def zero(self):
#        return self.num == 0
#
#    def uno(self):
#        return self.num == self.den
#
#    def menouno(self):
#        return self.num == -self.den

#FRA_TYPE = type(Fraction())

if __name__ == "__main__":
    ODT, TEX, PY, TXT = range(4)
    Fraction().out_format = TEX
    import doctest
    doctest.testmod(verbose=False)
